/*
	ProWesS, PROGS Windowing System

	ProWesS thing handling routines

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 9 January 1995
	18.09.1996 thing now always only used once by each job
	16.01.1997 DLL linkage has PWDoCreate added. this should only be linked
			   when using "v2" linkage.
*/

#include <thing.h>
#include <job.h>

#include "ProWesS.h"
#include "PWConfig.h"

extern char InitIsDone; 			/* to see if linking is safe to do */
extern Job ProWesS; 				/* job id for myself */

typedef struct {
	char identifier[8]; 			/* e.g. "<<QDLL>>" */
	void *usercode; 				/* base address for usercode */
	int extension;					/* name of extension in thing */
	short namelen;					/* length of thingname == 7 */
	char thingname[8];				/* "ProWesS " */
	char version[2];				/* "v1" or "v2" (PWDoCreate added) */
} DLLEntry;

#define PW_THING_CORE		0x434f5245	/* "CORE" */
#define PW_THING_PWS		0x50575320	/* "PWS " */
#define PW_THING_VERSION	0x312e3030	/* "1.00" */

static Error DLLink(DLLEntry *link)
{
	Error err;
	Job me, user;
	int counter, count;

	/* check version */
	if (link->version[0]!='v' || (link->version[1]!='1' && link->version[1]!='2'))
		return ERR_NIMP;

	JOBMe(&me);
retry:
	counter=0; count=0;
	while (!(err=THINGUserNext(PW_THING_NAME,&counter,&user)))
	{
		if (user==me)
		{
			count++;
			if (count>1) THINGFree(PW_THING_NAME);
		}
		if (counter==0) break;
	}
	if (err==ERR_IJOB) goto retry;

	if (me!=ProWesS)				/* I can always link back to myself */
	{
		/* wait until initialisation has finished */
		while (!InitIsDone)
			JOBSuspend(MYSELF,1,NULL);
	}

	/* now link the proper area */
	switch (link->extension)
	{
	case PW_THING_CORE:
		{
			extern char core_start[2], core_end[2];
			char *dummy=core_end;
			MEMCopy(dummy-core_start,core_start,link->usercode);
		}
		break;
	case PW_THING_PWS:
		{
			extern char pws_start[2], pws_end[2];
			char *dummy=pws_end;
			MEMCopy(dummy-pws_start-(link->version[1]=='2'?0:6),pws_start,link->usercode);
		}
		break;
	default:
		return ERR_IPAR;
	}
	return ERR_OK;
}

static ThingExtension thg_list[]={ { PW_THING_CORE, DLLink		 },
								   { PW_THING_PWS , DLLink       },
								   { PW_THING_CNFG, PWConfig     },
								   { NUL , (Error (*)(DLLEntry *))NULL }};

static ThingDefinition thg_defn={PW_THING_NAME, PW_THING_VERSION, FALSE,
								 NULL, NULL, NULL, NULL, 0, thg_list};

Error PWThingInit()
{
	Error err;
	Thing thing;

	/* remove old thing, unless still in use */
	err=THINGRemove(PW_THING_NAME);
	if (err && err!=ERR_ITNF) return err;

	/* link in new one */
	thg_defn.owner=ProWesS;
	err=THINGCreate(&thg_defn);
	if (err) return err;

	/* make sure the ProWesS job uses the ProWesS thing */
	return THINGUse(PW_THING_NAME,0,&thing,NULL);
}
